namespace BP
{
	class AI : BP::ScriptActor
	{
		float onFireTime; //[Serialize] time until fire caught on fire effects turn off
		bool isSpawned = false; //[Serialize] was custom spawned
		int lastOnDamageGameTick;
		bool editorLoad = false; //is loaded from the map and not spawned
		kActor@ movePoint; //a custom point to move to. Null if not moving to custom point
		kActor@ attackPoint; //a custom point to attack. Null if not attacking a custom point
		array<int> pathPoints; //list of TIDs to each pathpoint
		array<kActor@> pathPointActors;
		int pathPointIndex; //current path point ai is pathing to
		bool isPathMoving;
		bool pathIgnorePlayer;
		//------------------------------------------------------------------------------------------------------------------------
		AI(kActor @actor)
		{
			super(@actor);
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Called on AI that is in the map data not ones created with ActorFactory
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSpawn(void) override
		{
			editorLoad = true;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict) override
		{
			SERIALIZE_FLOAT(onFireTime);
			SERIALIZE_BOOL(isSpawned);
			SERIALIZE_BOOL(editorLoad);
			SERIALIZE_BOOL(isPathMoving);
			SERIALIZE_BOOL(pathIgnorePlayer);
			SERIALIZE_INT(pathPointIndex);
			
			//should define serializearray instead
			int pathPointsLength = int(pathPoints.length());
			SERIALIZE_INT(pathPointsLength);
			for (int i = 0; i < pathPointsLength; i++)
			{
				SERIALIZEKEY_INT("pathPoint" + i, pathPoints[i]);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict) override
		{
			DESERIALIZE_FLOAT(onFireTime);
			DESERIALIZE_BOOL(isSpawned);
			DESERIALIZE_BOOL(editorLoad);
			DESERIALIZE_BOOL(isPathMoving);
			DESERIALIZE_INT(pathPointIndex);
			DESERIALIZE_BOOL(pathIgnorePlayer);
			@movePoint = null;
			@attackPoint = null;
			
			//should define deserializearray instead
			int pathPointsLength;
			DESERIALIZE_INT(pathPointsLength);
			pathPoints.resize(pathPointsLength);
			for (int i = 0; i < pathPointsLength; i++)
			{
				DESERIALIZEKEY_INT("pathPoint" + i, pathPoints[i]);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnStart() override
		{
			//because spawned enemies default health to 100 set spawned enemies to definition start health
			if (!isSpawned && self.Definition() !is null && !editorLoad)
			{
				float health;
				self.Definition().GetFloat("health", health, 1.0f);
				self.Health() = health;
				if (IsMultiplayer())
				{
					self.Health() *= 2.0f; //enemies have double health in raptorfest mode
				}
			}
			BP::UserEvent::AIStarted(@BP::map, @self);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserializeStart() override
		{
			if (isPathMoving)
			{
				SetPath(pathPoints, pathIgnorePlayer, pathPointIndex);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
		{
			if (IsMultiplayer())
			{
				if (!BP::Actor::IsDead(@self))
				{
					kActor @src = dmgInfo.source;
					float realDamage = dmgInfo.hits;
					float moreHits = realDamage;
					if (src !is null && src.Type() == kActor_Player)
					{
						moreHits *= BP::playerDamage;
					}
					float hitsDif = moreHits - realDamage;
					self.Health() = MAX(self.Health() - hitsDif, realDamage);
					ShowDamage(dmgInfo, moreHits);
				}
			}		
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDamage(kDamageInfo& in dmgInfo) override
		{
			//Check if damaged with fire effect
			kActor @src = dmgInfo.source;
			if (@src != null)
			{
				bool hasFlameEffect;
				kDictMem @def = src.Definition();
				if (@def != null)
				{
					def.GetBool("BP_FlameEffect", hasFlameEffect);
					if (hasFlameEffect)
					{
						onFireTime = 4.0f;
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeath(kDamageInfo& in dmgInfo) override
		{
			BP::UserEvent::AIKilled(@BP::map, @self, @dmgInfo.source);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ShowDamage(kDamageInfo& in dmgInfo, float hits)
		{
			kActor @srcActor = dmgInfo.source;
			if (BP::showPlayerDamage && IsMultiplayer() && srcActor !is null)
			{
				float damage = hits;
				int damageDisplay = (damage > 0.0f && damage < 1.0f) ? Math::CeilToInt(damage) : int(damage);
				BP::Actor::ShowWorldText(@srcActor, @self, "" + damageDisplay);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_UserEvent(const int msg, const float x, const float y, const float z, const float f1, const float f2, const float f3) override
		{
			switch (msg)
			{
				case USEREVENT_MSG_SPAWN:
				{
					isSpawned = true;
					break;
				}
				case USEREVENT_MSG_AI_MOVE_POINT:
				{
					kVec3 point = BP::UserEvent::argVectors[0];
					MoveToPoint(point);
					break;
				}
				case USEREVENT_MSG_AI_ATTACK_POINT:
				{
					kVec3 point = BP::UserEvent::argVectors[0];
					AttackPoint(point);
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick() override
		{
			if (isPathMoving)
			{
				if (!BP::Actor::IsDead(@self))
				{
					kActor @sightTarget = @self.EnemyAIComponent().SightTarget().Target();
					if (!pathIgnorePlayer && (sightTarget !is null && sightTarget.Type() == kActor_Player))
					{
						//stop pathing
					}
					else //continue pathing
					{
						BP::Actor::AIIgnorePlayer(@self, pathIgnorePlayer);
						
						kActor @pathPointActor = @pathPointActors[pathPointIndex];
						float radius = pathPointActor.WorldComponent().Radius();
						//if reached path point go to next one
						if (BP::Actor::InRadius(@self, @pathPointActor, radius * radius))
						{
							pathPointIndex++;
							if (pathPointIndex >= int(pathPoints.length()))
							{
								pathPointIndex = 0;
							}
							@pathPointActor = @pathPointActors[pathPointIndex];
							BP::Actor::AIMoveToTarget(@self, @pathPointActor);
						}
					}
				}
			}
			
			if (onFireTime > 0.0f)
			{
				onFireTime -= GAME_DELTA_TIME;
				self.RenderMeshComponent().ToggleHotPoint(1, true);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void MoveToPoint(const kVec3&in point)
		{
			if (BP::Actor::Exists(@movePoint))
			{
				movePoint.WorldComponent().SetNearestPosition(point);
			}
			else
			{
				@movePoint = BP::Spawn::Actor(BP_Actor_WayPoint, point);
				movePoint.SetTarget(@self);
			}
			BP::Actor::AIMoveToTarget(@self, @movePoint);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void AttackPoint(const kVec3&in point)
		{
			if (BP::Actor::Exists(@attackPoint))
			{
				attackPoint.WorldComponent().SetNearestPosition(point);
			}
			else
			{
				@attackPoint = BP::Spawn::Actor(BP_Actor_WayPoint, point);
				attackPoint.SetTarget(@self);
			}
			BP::Actor::AIAttackTarget(@self, @attackPoint);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetPath(const array<int>& tagIDs, bool ignorePlayer, const int startPathIndex = 0)
		{
			pathPoints = tagIDs;
			pathPointIndex = startPathIndex;
			pathIgnorePlayer = ignorePlayer;
			isPathMoving = true;
			int pathPointsLength = int(pathPoints.length());
			pathPointActors.resize(pathPointsLength);
			if (pathPointsLength > 0)
			{
				for (int i = 0; i < pathPointsLength; i++)
				{
					@pathPointActors[i] = BP::Actor::FindActorWithTID(pathPoints[i]);
				}
				BP::Actor::AIMoveToTarget(@self, pathPointActors[pathPointIndex]);
			}			
			BP::Actor::AIIgnorePlayer(@self, pathIgnorePlayer);
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
